import tarfile
import warnings

import urllib
import ioh 
import numpy as np
import os
from .cluster_metrics import *
from .cluster_baseline_problems import *
from importlib.resources import files


def create_cluster_problem(dataset: str | np.ndarray, k: int, instance=1, error_metric="mse_euclidean") -> tuple[ioh.problem.RealSingleObjective, callable]:
    """
    Creates a clustering problem wrapped as an IOH problem for optimization.
    
    Parameters:
    ---------
        dataset : str or np.ndarray
            The dataset to be used for clustering. If a string is provided, it is assumed
            to be the name of a banchmark dataset. If an np.ndarray is provided, it is used directly
            as the dataset.
        k : int
            The number of clusters to create.
        instance : int, optional (default=1)
            The instance number for the problem. Useful for generating different problem
            instances with the same structure.
        error_metric : str or callable, optional (default="mse_euclidean")
            The error metric to evaluate the clustering. If a string is provided, it must
            match one of the predefined metrics in `CLUSTER_METRICS`. If a callable is provided,
            it should accept two arguments: the normalized dataset and the cluster centers.
    
    Returns:
    ---------
        f : ioh.problem.RealSingleObjective
            The IOH-wrapped clustering problem, ready for optimization.
        retransform : callable
                        A function that retransforms the optimized solution back to the original scale
                        of the dataset. It accepts a 1D array of cluster centers and returns a 2D array
                        of shape (k, number_of_features).

    Raises:
    ---------
        ValueError : If the provided `error_metric` is a string and does not match any predefined
                     metric in `CLUSTER_METRICS`.

    Notes:
        - The dataset is normalized to the range [0, 1] to ensure consistent scaling.
        - The `retransform` function can be used to map the normalized cluster centers
            back to their original scale.
        - The problem is set up for minimization, as clustering typically aims to minimize
            the error metric.
    """
    id = None
    if isinstance(dataset, str):
        id = get_problem_id(dataset)
        dataset_path = files("iohclustering.static").joinpath(f"{dataset}.txt")
        data = np.loadtxt(dataset_path, delimiter=',')

    else:
        data = dataset
        dataset = 'custom'

    if isinstance(error_metric, str):
        if error_metric in CLUSTER_METRICS.keys():
            error_metric = CLUSTER_METRICS[error_metric]
        else:
            raise ValueError(f"Unknown metric {error_metric}")
    

    data_np = np.array(data)
    # Normalize the data to the range [0, 1]
    data_min = np.tile(np.min(data_np, axis=0), k)
    data_max = np.tile(np.max(data_np, axis=0), k)
    data_normalized = (data_np - np.min(data_np, axis=0)) / (np.max(data_np, axis=0) - np.min(data_np, axis=0))
    
    # Define the objective function to be minimized
    def obj_func(x: np.ndarray) -> float:
        return error_metric(data_normalized, x.reshape(k, np.size(data_np, axis=1)))

    # Wrap the function as an IOH problem
    f = ioh.wrap_problem(
        obj_func,                                   # Handle to the function
        name=f"Cluster_{dataset}_k{k}",             # Name to be used when instantiating
        optimization_type=ioh.OptimizationType.MIN, # Specify that we want to minimize
        problem_class=ioh.ProblemClass.REAL,        # Problem class
        lb=0,                                       # Lower bound of the search space
        ub=1,                                       # Upper bound of the search space
        instance=instance,                          # Instance number
        dimension=int(k * np.size(data_np, axis=1)) # Dimension of the problem (k * number of features)
    )
    
    # Set the meta data for the problem
    if(id is not None):
        f.set_id(id)

    def retransform(X):
        X2 = ((X * (data_max - data_min)) + data_min)
        return X2.reshape(k, -1)

    return f, retransform


def get_problem_id(dataset_name: str) -> int:
    """
    Retrieve the problem ID associated with a given dataset name.
    This function takes a dataset name as input, converts it to lowercase, 
    and checks if it exists in the `CLUSTER_BASELINE_DATASETS` dictionary. 
    If the dataset name is found, it returns the corresponding problem ID. 
    Otherwise, it raises a `ValueError`.

    Parameters:
    ---------
        dataset_name (str): The name of the dataset to look up.
    
    Returns:
    ---------
        int: The problem ID associated with the given dataset name.
    
    Raises:
    ---------
        ValueError: If the dataset name is not found in `CLUSTER_BASELINE_DATASETS`.
    
    """
    dataset_name = dataset_name.lower()
    if dataset_name in CLUSTER_BASELINE_DATASETS.values():
        for k, v in CLUSTER_BASELINE_DATASETS.items():
            if v == dataset_name:
                return k
    else:
        raise ValueError(f"Unknown dataset name {dataset_name}")


def get_problem(fid: int | str, instance: int = 1, k: int = 2) -> ioh.problem.RealSingleObjective:
    """
    Retrieve a clustering problem based on the given dataset identifier.

    Parameters:
    ---------
        fid (int | str): The dataset identifier, either as an integer or a string.
                         If a string is provided, it will be converted to an integer
                         using the `get_problem_id` function.
        instance (int, optional): The instance of the dataset to use. Defaults to 1.
        k (int, optional): The number of clusters for the clustering problem. Defaults to 2.

    Returns:
    ---------
        f : ioh.problem.RealSingleObjective
              The IOH-wrapped clustering problem, ready for optimization.
        retransform : callable
                        A function that retransforms the optimized solution back to the original scale
                        of the dataset. It accepts a 1D array of cluster centers and returns a 2D array
                        of shape (k, number_of_features).

    Raises:
    ---------
        ValueError: If the provided dataset identifier is not found in the
                    `CLUSTER_BASELINE_DATASETS` dictionary.
    """
    if(type(fid) == str):
        fid = get_problem_id(fid)
    
    if fid not in CLUSTER_BASELINE_DATASETS.keys():
        raise ValueError(f"Unknown dataset id {fid}")

    return create_cluster_problem(
        dataset=CLUSTER_BASELINE_DATASETS[fid],
        k=k,
        instance=instance,
        error_metric=CLUSTER_METRICS["mse_euclidean"]
    )




def load_problems():
    """
    Loads clustering problems from the specified datasets path.

    This function iterates through a predefined set of baseline datasets and 
    checks if corresponding dataset files (with a `.txt` extension) exist in 
    the given directory. For each dataset, it generates clustering problems 
    for a set of predefined dimensions (k-values) and stores them in a dictionary.

    Parameters:
    ---------
        datasets_path (str): The path to the directory containing dataset files. 
                             Defaults to "banchmark_datasets".

    Returns:
    ---------
        dict: A dictionary where the keys are problem names (from metadata) 
              and the values are the corresponding clustering problem objects.

    """
    
    folder = files("iohclustering.static")
    datasets_names = [f.name for f in folder.iterdir() if f.suffix == ".txt"]

    problems = {}
    for dataset in CLUSTER_BASELINE_DATASETS.values():
        if f"{dataset}.txt" in datasets_names:
            for k in BASELINE_K_DIMENTIONS[dataset]:
                problem, retransform = create_cluster_problem(dataset, k)
                problems[problem.meta_data.name] = problem, retransform
    return problems




import unittest
import numpy as np
from iohclustering import mse_error

# filepath: src/IOHClustering/test_utils_metrics.py

class TestUtilsMetrics(unittest.TestCase):

    def test_mse_error_valid_input(self):
        # Test with a valid dataset, centroids, and labels
        X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
        centroids = np.array([[1, 2], [5, 6]])
        labels = np.array([0, 0, 1, 1])  # Assign first two points to cluster 0, last point to cluster 1
        mse = mse_error(X, centroids, labels)
        self.assertAlmostEqual(mse, 4.0, places=5)

    def test_mse_error_single_cluster(self):
        # Test with all points assigned to a single cluster
        X = np.array([[1, 2], [5, 6]])
        centroids = np.array([[3, 4]])
        labels = np.array([0, 0, 0])  # All points assigned to the single cluster
        mse = mse_error(X, centroids, labels)
        self.assertAlmostEqual(mse, 8.0, places=5)


if __name__ == "__main__":
    unittest.main()


import unittest
import numpy as np
from unittest.mock import patch, MagicMock

# filepath: src/IOHClustering/test_cluster_metrics.py
from iohclustering import (
    euclidean_distance, 
    mse_error,
    compute_labels,
    general_cluster_metric,
    CLUSTER_METRICS,
)


class TestClusterMetrics(unittest.TestCase):

    def test_compute_labels(self):
        X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
        centroids = np.array([[1, 2], [5, 6]])
        labels = compute_labels(X, centroids, euclidean_distance)
        self.assertTrue(np.array_equal(labels, [0, 0, 1, 1]))

    def test_general_cluster_metric(self):
        X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
        centroids = np.array([[1, 2], [5, 6]])

        metric_function = general_cluster_metric(euclidean_distance, mse_error)
        result = metric_function(X, centroids)

        self.assertEqual(result, 4.0)

    def test_general_cluster_metric_default(self):
        metric_function = general_cluster_metric()
        X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
        centroids = np.array([[1, 2], [5, 6]])
        result = metric_function(X, centroids)
        self.assertGreaterEqual(result, 0)
        self.assertAlmostEqual(result, 4.0)

    def test_cluster_metrics_dict(self):
        self.assertIn("mse_euclidean", CLUSTER_METRICS)
        metric_function = CLUSTER_METRICS["mse_euclidean"]
        X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
        centroids = np.array([[1, 2], [5, 6]])
        result = metric_function(X, centroids)
        self.assertGreaterEqual(result, 0)
        self.assertAlmostEqual(result, 4.0)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import typing
from .utils_metrics import euclidean_distance, mse_error

def compute_labels(
        X: np.ndarray, 
        centroids: np.ndarray, 
        distance_function: typing.Callable = euclidean_distance
    ):
    """
    Compute the labels for each data point based on the closest centroid.

    Parameters:
    - X: numpy array of shape (m, n), where m is the number of data points and n is the dimensionality of the data.
    - centroids: numpy array of shape (k, n), where k is the number of clusters and n is the dimensionality of the centroids.

    Returns:
    - labels: numpy array of shape (m,), where each element is the index of the closest centroid for each data point.
    """
    # Initialize an array to store the labels (cluster assignments)
    labels = np.zeros(len(X), dtype=int)

    # Calculate labels (assign closest centroid to each data point)
    for i in range(len(X)):
        # Compute the distance between the point and each centroid
        distances = distance_function(X[i], centroids)
        # Assign the label as the index of the closest centroid
        labels[i] = np.argmin(distances)

    return labels


def general_cluster_metric(
    distance_function: typing.Callable[[np.ndarray, np.ndarray], np.ndarray] = euclidean_distance, 
    error_function: typing.Callable[[np.ndarray, np.ndarray, np.ndarray[np.int_]], float] = mse_error
    ) -> typing.Callable[[np.ndarray, np.ndarray], float]:
    """
    Creates a general clustering metric function based on the provided distance and error functions.

    Parameters:
    --------
        distance_function (Callable, optional): A function to compute the distance between 
            data points and centroids. Defaults to `euclidean_distance`.
        error_function (Callable, optional): A function to compute the clustering error 
            based on the dataset, centroids, and labels. Defaults to `mse_error`.

    Returns:
    --------
        Callable[[np.ndarray, np.ndarray], float]: A function that computes the clustering metric.
    """
    def metric(X: np.ndarray, centroids: np.ndarray) -> float:
        labels = compute_labels(X, centroids, distance_function)
        return error_function(X, centroids, labels)
    
    return metric



CLUSTER_METRICS = {
    "mse_euclidean": general_cluster_metric(euclidean_distance, mse_error),
}



CLUSTER_BASELINE_DATASETS = {
    1: "breast_pca",
    2: "diabetes_pca",
    3: "german_postal_selected",
    4: "glass_pca",
    5: "iris_pca",
    6: "kc1_pca",
    7: "mfeat-fourier_pca",
    8: "ruspini_selected",
    9: "segment_pca",
    10: "wine_pca",
}

BASELINE_K_DIMENTIONS={
    "breast_pca": [2, 3, 5, 10], 
    "diabetes_pca": [2, 3, 5, 10],
    "german_postal_selected": [2, 3, 5, 10],
    "glass_pca": [2, 3, 5, 10],
    "iris_pca": [2, 3, 5, 10],
    "kc1_pca": [2, 3, 5, 10],
    "mfeat-fourier_pca": [2, 3, 5, 10],
    "ruspini_selected": [2, 3, 5, 10],
    "segment_pca": [2, 3, 5, 10],
    "wine_pca": [2, 3, 5, 10],
}



import unittest
from unittest.mock import patch, MagicMock
import numpy as np
import ioh
from iohclustering.cluster_base import create_cluster_problem, get_problem_id, get_problem, load_problems


class TestClusterBase(unittest.TestCase):

    def test_load_problems(self):
        # Test loading problems
        problems = load_problems()
        self.assertIsInstance(problems, dict)
        self.assertGreater(len(problems), 0)
        self.assertEqual(len(problems), 40)
        
        from iohclustering.cluster_baseline_problems import CLUSTER_BASELINE_DATASETS, BASELINE_K_DIMENTIONS
        # Check if the problems loaded match the expected datasets and dimensions
        for dataset, k_values in BASELINE_K_DIMENTIONS.items():
            for k in k_values:
                problem_name = f"Cluster_{dataset}_k{k}"
                self.assertIn(problem_name, problems.keys())

    def test_get_problem_id_valid(self):
        # Test valid dataset names
        self.assertEqual(get_problem_id("breast_pca"), 1)
        self.assertEqual(get_problem_id("bReAsT_PcA"), 1)  # Test case-insensitivity

    def test_get_problem_id_invalid(self):
        # Test invalid dataset name
        with self.assertRaises(ValueError) as context:
            get_problem_id("unknown_dataset")
        self.assertEqual(str(context.exception), "Unknown dataset name unknown_dataset")

    def test_create_cluster_problem(self):

        # Test creating a clustering problem
        k = 2
        f, retransform = create_cluster_problem("breast_pca", k= k, instance=1, error_metric="mse_euclidean")
        self.assertEqual(f.meta_data.name, "Cluster_breast_pca_k2")
        self.assertEqual(f.meta_data.problem_id, 1)
        self.assertEqual(f.meta_data.instance, 1)
        self.assertEqual(f.meta_data.n_variables, k*2)
        self.assertEqual(f.meta_data.optimization_type, ioh.OptimizationType.MIN)

    def test_create_cluster_problem_custom_data(self):

        # Test creating a clustering problem
        data = np.array([[0,0], [1,1], [3,3], [4,4]])
        k = 2
        f, retransform = create_cluster_problem(data, k= k, instance=1, error_metric="mse_euclidean")
        self.assertEqual(f.meta_data.name, "Cluster_custom_k2")
        self.assertEqual(f.meta_data.instance, 1)
        self.assertEqual(f.meta_data.n_variables, k*2)
        self.assertEqual(f.meta_data.optimization_type, ioh.OptimizationType.MIN)

        input_vec = np.array([[1.0, 2.0],[3.0, 4.0]])
        transformed_input_vec = np.array([0.25, 0.5, 0.75, 1.0])
        self.assertTrue(np.allclose(retransform(transformed_input_vec), input_vec))

    def test_get_problem(self):
        k = 2
        f, retransform = get_problem(1, instance=1, k=k)
        self.assertEqual(f.meta_data.name, "Cluster_breast_pca_k2")
        self.assertEqual(f.meta_data.problem_id, 1)
        self.assertEqual(f.meta_data.instance, 1)
        self.assertEqual(f.meta_data.n_variables, k*2)
        self.assertEqual(f.meta_data.optimization_type, ioh.OptimizationType.MIN)

        # Test retrieving a problem by dataset name
        f, retransform = get_problem("breast_pca", instance=1, k=2)
        self.assertEqual(f.meta_data.name, "Cluster_breast_pca_k2")
        self.assertEqual(f.meta_data.problem_id, 1)
        self.assertEqual(f.meta_data.instance, 1)
        self.assertEqual(f.meta_data.n_variables, k*2)
        self.assertEqual(f.meta_data.optimization_type, ioh.OptimizationType.MIN)

        # Test invalid dataset ID
        with self.assertRaises(ValueError) as context:
            get_problem(99, instance=1, k=2)
        self.assertEqual(str(context.exception), "Unknown dataset id 99")

    



if __name__ == "__main__":
    unittest.main()


import numpy as np

def mse_error(X: np.ndarray, centroids: np.ndarray, labels: np.ndarray) -> float:
    """
    Compute the Mean Squared Error (MSE) for a clustering problem.
    The MSE is calculated as the average of the squared Euclidean distances 
    between each data point and the centroid of its assigned cluster.
    Parameters:
    ----------
        X : np.ndarray
            A 2D array of shape (n_samples, n_features) representing the dataset.
        centroids : np.ndarray
            A 2D array of shape (n_clusters, n_features) representing the cluster centroids.
        labels : np.ndarray
            A 1D array of shape (n_samples,) where each element is the index of the 
            cluster assigned to the corresponding data point in X.
    Returns:
    -------
        float
            The mean squared error of the clustering.
    """
    
    
    mse = 0
    for i in range(len(X)):
        # Get the centroid of the assigned cluster
        centroid = centroids[labels[i]]
        # Compute the squared error manually (Euclidean distance squared)
        squared_error = np.sum((X[i] - centroid) ** 2)
        mse += squared_error

    # Average the MSE by the number of data points
    mse /= len(X)

    return mse

def euclidean_distance(x: np.ndarray, centroids: np.ndarray) -> np.ndarray:
    """
    Compute the Euclidean distance between a point and a set of centroids.

    Parameters:
    ---------
        x (numpy.ndarray): A 1D array representing the point.
        centroids (numpy.ndarray): A 2D array where each row represents a centroid.

    Returns:
    --------
        numpy.ndarray: A 1D array containing the Euclidean distances from the point
                   to each centroid.
    """
    return np.linalg.norm(x - centroids, axis=1)